@@ -21,20 +21,16 @@
 
 /**
  * SECTION:element-pulsesrc
- * @short_description: Capture audio from a PulseAudio sound server
  * @see_also: pulsesink, pulsemixer
  *
+ * This element captures audio from a
+ * <ulink href="http://www.pulseaudio.org">PulseAudio sound server</ulink>.
+ *
  * <refsect2>
- * <para>
- * This element captures audio from a PulseAudio sound server.
- * </para>
  * <title>Example pipelines</title>
- * <para>
- * <programlisting>
+ * |[
  * gst-launch -v pulsesrc ! audioconvert ! vorbisenc ! oggmux ! filesink location=alsasrc.ogg
- * </programlisting>
- * Record from a sound card using ALSA and encode to Ogg/Vorbis.
- * </para>
+ * ]| Record from a sound card using pulseaudio and encode to Ogg/Vorbis.
  * </refsect2>
  */
 
@@ -55,11 +51,17 @@
 GST_DEBUG_CATEGORY_EXTERN (pulse_debug);
 #define GST_CAT_DEFAULT pulse_debug
 
+#define DEFAULT_SERVER            NULL
+#define DEFAULT_DEVICE            NULL
+#define DEFAULT_DEVICE_NAME       NULL
+
 enum
 {
-  PROP_SERVER = 1,
+  PROP_0,
+  PROP_SERVER,
   PROP_DEVICE,
-  PROP_DEVICE_NAME
+  PROP_DEVICE_NAME,
+  PROP_LAST
 };
 
 static void gst_pulsesrc_destroy_stream (GstPulseSrc * pulsesrc);
@@ -72,8 +74,6 @@ static void gst_pulsesrc_get_property (GObject * object, guint prop_id,
     GValue * value, GParamSpec * pspec);
 static void gst_pulsesrc_finalize (GObject * object);
 
-static void gst_pulsesrc_dispose (GObject * object);
-
 static gboolean gst_pulsesrc_open (GstAudioSrc * asrc);
 
 static gboolean gst_pulsesrc_close (GstAudioSrc * asrc);
@@ -87,6 +87,8 @@ static guint gst_pulsesrc_read (GstAudioSrc * asrc, gpointer data,
     guint length);
 static guint gst_pulsesrc_delay (GstAudioSrc * asrc);
 
+static void gst_pulsesrc_reset (GstAudioSrc * src);
+
 static gboolean gst_pulsesrc_negotiate (GstBaseSrc * basesrc);
 
 static GstStateChangeReturn gst_pulsesrc_change_state (GstElement *
@@ -109,7 +111,7 @@ static gboolean
 gst_pulsesrc_interface_supported (GstImplementsInterface *
     iface, GType interface_type)
 {
-  GstPulseSrc *this = GST_PULSESRC (iface);
+  GstPulseSrc *this = GST_PULSESRC_CAST (iface);
 
   if (interface_type == GST_TYPE_MIXER && this->mixer)
     return TRUE;
@@ -165,30 +167,30 @@ gst_pulsesrc_base_init (gpointer g_class)
           "width = (int) 16, "
           "depth = (int) 16, "
           "rate = (int) [ 1, MAX ], "
-          "channels = (int) [ 1, 16 ];"
+          "channels = (int) [ 1, 32 ];"
+          "audio/x-raw-float, "
+          "endianness = (int) { " ENDIANNESS " }, "
+          "width = (int) 32, "
+          "rate = (int) [ 1, MAX ], "
+          "channels = (int) [ 1, 32 ];"
           "audio/x-raw-int, "
           "endianness = (int) { " ENDIANNESS " }, "
           "signed = (boolean) TRUE, "
           "width = (int) 32, "
           "depth = (int) 32, "
           "rate = (int) [ 1, MAX ], "
-          "channels = (int) [ 1, 16 ];"
-          "audio/x-raw-float, "
-          "endianness = (int) { " ENDIANNESS " }, "
-          "width = (int) 32, "
-          "rate = (int) [ 1, MAX ], "
-          "channels = (int) [ 1, 16 ];"
+          "channels = (int) [ 1, 32 ];"
           "audio/x-raw-int, "
           "signed = (boolean) FALSE, "
           "width = (int) 8, "
           "depth = (int) 8, "
           "rate = (int) [ 1, MAX ], "
-          "channels = (int) [ 1, 16 ];"
+          "channels = (int) [ 1, 32 ];"
           "audio/x-alaw, "
           "rate = (int) [ 1, MAX], "
-          "channels = (int) [ 1, 16 ];"
+          "channels = (int) [ 1, 32 ];"
           "audio/x-mulaw, "
-          "rate = (int) [ 1, MAX], " "channels = (int) [ 1, 16 ]")
+          "rate = (int) [ 1, MAX], " "channels = (int) [ 1, 32 ]")
       );
 
   GstElementClass *element_class = GST_ELEMENT_CLASS (g_class);
@@ -209,14 +211,13 @@ gst_pulsesrc_class_init (GstPulseSrcClass * klass)
   GstBaseSrcClass *gstbasesrc_class = GST_BASE_SRC_CLASS (klass);
   GstElementClass *gstelement_class = GST_ELEMENT_CLASS (klass);
 
-  gstelement_class->change_state =
-      GST_DEBUG_FUNCPTR (gst_pulsesrc_change_state);
-
-  gobject_class->dispose = GST_DEBUG_FUNCPTR (gst_pulsesrc_dispose);
   gobject_class->finalize = GST_DEBUG_FUNCPTR (gst_pulsesrc_finalize);
   gobject_class->set_property = GST_DEBUG_FUNCPTR (gst_pulsesrc_set_property);
   gobject_class->get_property = GST_DEBUG_FUNCPTR (gst_pulsesrc_get_property);
 
+  gstelement_class->change_state =
+      GST_DEBUG_FUNCPTR (gst_pulsesrc_change_state);
+
   gstbasesrc_class->negotiate = GST_DEBUG_FUNCPTR (gst_pulsesrc_negotiate);
 
   gstaudiosrc_class->open = GST_DEBUG_FUNCPTR (gst_pulsesrc_open);
@@ -225,22 +226,24 @@ gst_pulsesrc_class_init (GstPulseSrcClass * klass)
   gstaudiosrc_class->unprepare = GST_DEBUG_FUNCPTR (gst_pulsesrc_unprepare);
   gstaudiosrc_class->read = GST_DEBUG_FUNCPTR (gst_pulsesrc_read);
   gstaudiosrc_class->delay = GST_DEBUG_FUNCPTR (gst_pulsesrc_delay);
+  gstaudiosrc_class->reset = GST_DEBUG_FUNCPTR (gst_pulsesrc_reset);
 
   /* Overwrite GObject fields */
   g_object_class_install_property (gobject_class,
       PROP_SERVER,
       g_param_spec_string ("server", "Server",
-          "The PulseAudio server to connect to", NULL,
+          "The PulseAudio server to connect to", DEFAULT_SERVER,
           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+
   g_object_class_install_property (gobject_class, PROP_DEVICE,
-      g_param_spec_string ("device", "Source",
-          "The PulseAudio source device to connect to", NULL,
+      g_param_spec_string ("device", "Device",
+          "The PulseAudio source device to connect to", DEFAULT_DEVICE,
           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
 
   g_object_class_install_property (gobject_class,
       PROP_DEVICE_NAME,
       g_param_spec_string ("device-name", "Device name",
-          "Human-readable name of the sound device", NULL,
+          "Human-readable name of the sound device", DEFAULT_DEVICE_NAME,
           G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
 }
 
@@ -249,7 +252,9 @@ gst_pulsesrc_init (GstPulseSrc * pulsesrc, GstPulseSrcClass * klass)
 {
   int e;
 
-  pulsesrc->server = pulsesrc->device = NULL;
+  pulsesrc->server = NULL;
+  pulsesrc->device = NULL;
+  pulsesrc->device_description = NULL;
 
   pulsesrc->context = NULL;
   pulsesrc->stream = NULL;
@@ -257,6 +262,18 @@ gst_pulsesrc_init (GstPulseSrc * pulsesrc, GstPulseSrcClass * klass)
   pulsesrc->read_buffer = NULL;
   pulsesrc->read_buffer_length = 0;
 
+#if HAVE_PULSE_0_9_13
+  pa_sample_spec_init (&pulsesrc->sample_spec);
+#else
+  pulsesrc->sample_spec.format = PA_SAMPLE_INVALID;
+  pulsesrc->sample_spec.rate = 0;
+  pulsesrc->sample_spec.channels = 0;
+#endif
+
+  pulsesrc->operation_success = FALSE;
+  pulsesrc->paused = FALSE;
+  pulsesrc->in_read = FALSE;
+
   pulsesrc->mainloop = pa_threaded_mainloop_new ();
   g_assert (pulsesrc->mainloop);
 
@@ -265,7 +282,11 @@ gst_pulsesrc_init (GstPulseSrc * pulsesrc, GstPulseSrcClass * klass)
 
   pulsesrc->mixer = NULL;
 
-  pulsesrc->probe = gst_pulseprobe_new (G_OBJECT (pulsesrc), G_OBJECT_GET_CLASS (pulsesrc), PROP_DEVICE, pulsesrc->device, FALSE, TRUE);        /* FALSE for sinks, TRUE for sources */
+  pulsesrc->probe = gst_pulseprobe_new (G_OBJECT (pulsesrc), G_OBJECT_GET_CLASS (pulsesrc), PROP_DEVICE, pulsesrc->server, FALSE, TRUE);        /* FALSE for sinks, TRUE for sources */
+
+  /* this should be the default but it isn't yet */
+  gst_base_audio_src_set_slave_method (GST_BASE_AUDIO_SRC (pulsesrc),
+      GST_BASE_AUDIO_SRC_SLAVE_SKEW);
 }
 
 static void
@@ -276,6 +297,9 @@ gst_pulsesrc_destroy_stream (GstPulseSrc * pulsesrc)
     pa_stream_unref (pulsesrc->stream);
     pulsesrc->stream = NULL;
   }
+
+  g_free (pulsesrc->device_description);
+  pulsesrc->device_description = NULL;
 }
 
 static void
@@ -294,7 +318,7 @@ gst_pulsesrc_destroy_context (GstPulseSrc * pulsesrc)
 static void
 gst_pulsesrc_finalize (GObject * object)
 {
-  GstPulseSrc *pulsesrc = GST_PULSESRC (object);
+  GstPulseSrc *pulsesrc = GST_PULSESRC_CAST (object);
 
   pa_threaded_mainloop_stop (pulsesrc->mainloop);
 
@@ -305,8 +329,10 @@ gst_pulsesrc_finalize (GObject * object)
 
   pa_threaded_mainloop_free (pulsesrc->mainloop);
 
-  if (pulsesrc->mixer)
+  if (pulsesrc->mixer) {
     gst_pulsemixer_ctrl_free (pulsesrc->mixer);
+    pulsesrc->mixer = NULL;
+  }
 
   if (pulsesrc->probe) {
     gst_pulseprobe_free (pulsesrc->probe);
@@ -316,65 +342,126 @@ gst_pulsesrc_finalize (GObject * object)
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
-static void
-gst_pulsesrc_dispose (GObject * object)
+static gboolean
+gst_pulsesrc_is_dead (GstPulseSrc * pulsesrc)
 {
-  G_OBJECT_CLASS (parent_class)->dispose (object);
+
+  if (!pulsesrc->context
+      || !PA_CONTEXT_IS_GOOD (pa_context_get_state (pulsesrc->context))
+      || !pulsesrc->stream
+      || !PA_STREAM_IS_GOOD (pa_stream_get_state (pulsesrc->stream))) {
+    const gchar *err_str = pulsesrc->context ?
+        pa_strerror (pa_context_errno (pulsesrc->context)) : NULL;
+
+    GST_ELEMENT_ERROR ((pulsesrc), RESOURCE, FAILED, ("Disconnected: %s",
+            err_str), (NULL));
+    return TRUE;
+  }
+
+  return FALSE;
 }
 
+
 static void
 gst_pulsesrc_set_property (GObject * object,
     guint prop_id, const GValue * value, GParamSpec * pspec)
 {
 
-  GstPulseSrc *pulsesrc = GST_PULSESRC (object);
+  GstPulseSrc *pulsesrc = GST_PULSESRC_CAST (object);
 
   switch (prop_id) {
     case PROP_SERVER:
       g_free (pulsesrc->server);
       pulsesrc->server = g_value_dup_string (value);
-
       if (pulsesrc->probe)
         gst_pulseprobe_set_server (pulsesrc->probe, pulsesrc->server);
-
       break;
-
     case PROP_DEVICE:
       g_free (pulsesrc->device);
       pulsesrc->device = g_value_dup_string (value);
       break;
-
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
   }
 }
 
+static void
+gst_pulsesrc_source_info_cb (pa_context * c, const pa_source_info * i, int eol,
+    void *userdata)
+{
+  GstPulseSrc *pulsesrc = GST_PULSESRC_CAST (userdata);
+
+  if (!i)
+    return;
+
+  if (!pulsesrc->stream)
+    return;
+
+  g_assert (i->index == pa_stream_get_device_index (pulsesrc->stream));
+
+  g_free (pulsesrc->device_description);
+  pulsesrc->device_description = g_strdup (i->description);
+}
+
+static gchar *
+gst_pulsesrc_device_description (GstPulseSrc * pulsesrc)
+{
+  pa_operation *o = NULL;
+  gchar *t;
+
+  pa_threaded_mainloop_lock (pulsesrc->mainloop);
+
+  if (!pulsesrc->stream)
+    goto unlock;
+
+  if (!(o = pa_context_get_source_info_by_index (pulsesrc->context,
+              pa_stream_get_device_index (pulsesrc->stream),
+              gst_pulsesrc_source_info_cb, pulsesrc))) {
+
+    GST_ELEMENT_ERROR (pulsesrc, RESOURCE, FAILED,
+        ("pa_stream_get_source_info() failed: %s",
+            pa_strerror (pa_context_errno (pulsesrc->context))), (NULL));
+    goto unlock;
+  }
+
+  while (pa_operation_get_state (o) == PA_OPERATION_RUNNING) {
+
+    if (gst_pulsesrc_is_dead (pulsesrc))
+      goto unlock;
+
+    pa_threaded_mainloop_wait (pulsesrc->mainloop);
+  }
+
+unlock:
+
+  if (o)
+    pa_operation_unref (o);
+
+  t = g_strdup (pulsesrc->device_description);
+
+  pa_threaded_mainloop_unlock (pulsesrc->mainloop);
+
+  return t;
+}
+
 static void
 gst_pulsesrc_get_property (GObject * object,
     guint prop_id, GValue * value, GParamSpec * pspec)
 {
 
-  GstPulseSrc *pulsesrc = GST_PULSESRC (object);
+  GstPulseSrc *pulsesrc = GST_PULSESRC_CAST (object);
 
   switch (prop_id) {
     case PROP_SERVER:
       g_value_set_string (value, pulsesrc->server);
       break;
-
     case PROP_DEVICE:
       g_value_set_string (value, pulsesrc->device);
       break;
-
     case PROP_DEVICE_NAME:
-
-      if (pulsesrc->mixer)
-        g_value_set_string (value, pulsesrc->mixer->description);
-      else
-        g_value_set_string (value, NULL);
-
+      g_value_take_string (value, gst_pulsesrc_device_description (pulsesrc));
       break;
-
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -384,7 +471,7 @@ gst_pulsesrc_get_property (GObject * object,
 static void
 gst_pulsesrc_context_state_cb (pa_context * c, void *userdata)
 {
-  GstPulseSrc *pulsesrc = GST_PULSESRC (userdata);
+  GstPulseSrc *pulsesrc = GST_PULSESRC_CAST (userdata);
 
   switch (pa_context_get_state (c)) {
     case PA_CONTEXT_READY:
@@ -404,7 +491,7 @@ gst_pulsesrc_context_state_cb (pa_context * c, void *userdata)
 static void
 gst_pulsesrc_stream_state_cb (pa_stream * s, void *userdata)
 {
-  GstPulseSrc *pulsesrc = GST_PULSESRC (userdata);
+  GstPulseSrc *pulsesrc = GST_PULSESRC_CAST (userdata);
 
   switch (pa_stream_get_state (s)) {
 
@@ -423,20 +510,61 @@ gst_pulsesrc_stream_state_cb (pa_stream * s, void *userdata)
 static void
 gst_pulsesrc_stream_request_cb (pa_stream * s, size_t length, void *userdata)
 {
-  GstPulseSrc *pulsesrc = GST_PULSESRC (userdata);
+  GstPulseSrc *pulsesrc = GST_PULSESRC_CAST (userdata);
 
-  pa_threaded_mainloop_signal (pulsesrc->mainloop, 0);
+  GST_LOG_OBJECT (pulsesrc, "got request for length %" G_GSIZE_FORMAT, length);
+
+  if (pulsesrc->in_read) {
+    /* only signal when reading */
+    pa_threaded_mainloop_signal (pulsesrc->mainloop, 0);
+  }
+}
+
+static void
+gst_pulsesrc_stream_latency_update_cb (pa_stream * s, void *userdata)
+{
+  const pa_timing_info *info;
+  pa_usec_t source_usec;
+
+  info = pa_stream_get_timing_info (s);
+
+#if HAVE_PULSE_0_9_11
+  source_usec = info->configured_source_usec;
+#else
+  source_usec = 0;
+#endif
+
+  GST_LOG_OBJECT (GST_PULSESRC_CAST (userdata),
+      "latency_update, %" G_GUINT64_FORMAT ", %d:%" G_GINT64_FORMAT ", %d:%"
+      G_GUINT64_FORMAT ", %" G_GUINT64_FORMAT ", %" G_GUINT64_FORMAT,
+      GST_TIMEVAL_TO_TIME (info->timestamp), info->write_index_corrupt,
+      info->write_index, info->read_index_corrupt, info->read_index,
+      info->source_usec, source_usec);
+}
+
+static void
+gst_pulsesrc_stream_underflow_cb (pa_stream * s, void *userdata)
+{
+  GST_WARNING_OBJECT (GST_PULSESRC_CAST (userdata), "Got underflow");
+}
+
+static void
+gst_pulsesrc_stream_overflow_cb (pa_stream * s, void *userdata)
+{
+  GST_WARNING_OBJECT (GST_PULSESRC_CAST (userdata), "Got overflow");
 }
 
 static gboolean
 gst_pulsesrc_open (GstAudioSrc * asrc)
 {
-  GstPulseSrc *pulsesrc = GST_PULSESRC (asrc);
-
+  GstPulseSrc *pulsesrc = GST_PULSESRC_CAST (asrc);
   gchar *name = gst_pulse_client_name ();
 
   pa_threaded_mainloop_lock (pulsesrc->mainloop);
 
+  g_assert (!pulsesrc->context);
+  g_assert (!pulsesrc->stream);
+
   if (!(pulsesrc->context =
           pa_context_new (pa_threaded_mainloop_get_api (pulsesrc->mainloop),
               name))) {
@@ -454,13 +582,22 @@ gst_pulsesrc_open (GstAudioSrc * asrc)
     goto unlock_and_fail;
   }
 
-  /* Wait until the context is ready */
-  pa_threaded_mainloop_wait (pulsesrc->mainloop);
+  for (;;) {
+    pa_context_state_t state;
 
-  if (pa_context_get_state (pulsesrc->context) != PA_CONTEXT_READY) {
-    GST_ELEMENT_ERROR (pulsesrc, RESOURCE, FAILED, ("Failed to connect: %s",
-            pa_strerror (pa_context_errno (pulsesrc->context))), (NULL));
-    goto unlock_and_fail;
+    state = pa_context_get_state (pulsesrc->context);
+
+    if (!PA_CONTEXT_IS_GOOD (state)) {
+      GST_ELEMENT_ERROR (pulsesrc, RESOURCE, FAILED, ("Failed to connect: %s",
+              pa_strerror (pa_context_errno (pulsesrc->context))), (NULL));
+      goto unlock_and_fail;
+    }
+
+    if (state == PA_CONTEXT_READY)
+      break;
+
+    /* Wait until the context is ready */
+    pa_threaded_mainloop_wait (pulsesrc->mainloop);
   }
 
   pa_threaded_mainloop_unlock (pulsesrc->mainloop);
@@ -468,18 +605,22 @@ gst_pulsesrc_open (GstAudioSrc * asrc)
   g_free (name);
   return TRUE;
 
+  /* ERRORS */
 unlock_and_fail:
+  {
+    gst_pulsesrc_destroy_context (pulsesrc);
 
-  pa_threaded_mainloop_unlock (pulsesrc->mainloop);
+    pa_threaded_mainloop_unlock (pulsesrc->mainloop);
 
-  g_free (name);
-  return FALSE;
+    g_free (name);
+    return FALSE;
+  }
 }
 
 static gboolean
 gst_pulsesrc_close (GstAudioSrc * asrc)
 {
-  GstPulseSrc *pulsesrc = GST_PULSESRC (asrc);
+  GstPulseSrc *pulsesrc = GST_PULSESRC_CAST (asrc);
 
   pa_threaded_mainloop_lock (pulsesrc->mainloop);
   gst_pulsesrc_destroy_context (pulsesrc);
@@ -491,7 +632,7 @@ gst_pulsesrc_close (GstAudioSrc * asrc)
 static gboolean
 gst_pulsesrc_unprepare (GstAudioSrc * asrc)
 {
-  GstPulseSrc *pulsesrc = GST_PULSESRC (asrc);
+  GstPulseSrc *pulsesrc = GST_PULSESRC_CAST (asrc);
 
   pa_threaded_mainloop_lock (pulsesrc->mainloop);
   gst_pulsesrc_destroy_stream (pulsesrc);
@@ -504,49 +645,51 @@ gst_pulsesrc_unprepare (GstAudioSrc * asrc)
   return TRUE;
 }
 
-#define CHECK_DEAD_GOTO(pulsesrc, label) \
-if (!(pulsesrc)->context || pa_context_get_state((pulsesrc)->context) != PA_CONTEXT_READY || \
-    !(pulsesrc)->stream || pa_stream_get_state((pulsesrc)->stream) != PA_STREAM_READY) { \
-    GST_ELEMENT_ERROR((pulsesrc), RESOURCE, FAILED, ("Disconnected: %s", (pulsesrc)->context ? pa_strerror(pa_context_errno((pulsesrc)->context)) : NULL), (NULL)); \
-    goto label; \
-}
-
 static guint
 gst_pulsesrc_read (GstAudioSrc * asrc, gpointer data, guint length)
 {
-  GstPulseSrc *pulsesrc = GST_PULSESRC (asrc);
-
+  GstPulseSrc *pulsesrc = GST_PULSESRC_CAST (asrc);
   size_t sum = 0;
 
   pa_threaded_mainloop_lock (pulsesrc->mainloop);
+  pulsesrc->in_read = TRUE;
 
-  CHECK_DEAD_GOTO (pulsesrc, unlock_and_fail);
+  if (pulsesrc->paused)
+    goto was_paused;
 
   while (length > 0) {
     size_t l;
 
+    GST_LOG_OBJECT (pulsesrc, "reading %u bytes", length);
+
+    /*check if we have a leftover buffer */
     if (!pulsesrc->read_buffer) {
-
       for (;;) {
-        if (pa_stream_peek (pulsesrc->stream, &pulsesrc->read_buffer,
-                &pulsesrc->read_buffer_length) < 0) {
-          GST_ELEMENT_ERROR (pulsesrc, RESOURCE, FAILED,
-              ("pa_stream_peek() failed: %s",
-                  pa_strerror (pa_context_errno (pulsesrc->context))), (NULL));
+        if (gst_pulsesrc_is_dead (pulsesrc))
           goto unlock_and_fail;
-        }
 
-        if (pulsesrc->read_buffer)
+        /* read all available data, we keep a pointer to the data and the length
+         * and take from it what we need. */
+        if (pa_stream_peek (pulsesrc->stream, &pulsesrc->read_buffer,
+                &pulsesrc->read_buffer_length) < 0)
+          goto peek_failed;
+
+        GST_LOG_OBJECT (pulsesrc, "have data of %" G_GSIZE_FORMAT " bytes",
+            pulsesrc->read_buffer_length);
+
+        /* if we have data, process if */
+        if (pulsesrc->read_buffer && pulsesrc->read_buffer_length)
           break;
 
+        /* now wait for more data to become available */
+        GST_LOG_OBJECT (pulsesrc, "waiting for data");
         pa_threaded_mainloop_wait (pulsesrc->mainloop);
 
-        CHECK_DEAD_GOTO (pulsesrc, unlock_and_fail);
+        if (pulsesrc->paused)
+          goto was_paused;
       }
     }
 
-    g_assert (pulsesrc->read_buffer && pulsesrc->read_buffer_length);
-
     l = pulsesrc->read_buffer_length >
         length ? length : pulsesrc->read_buffer_length;
 
@@ -557,70 +700,90 @@ gst_pulsesrc_read (GstAudioSrc * asrc, gpointer data, guint length)
 
     data = (guint8 *) data + l;
     length -= l;
-
     sum += l;
 
     if (pulsesrc->read_buffer_length <= 0) {
+      /* we copied all of the data, drop it now */
+      if (pa_stream_drop (pulsesrc->stream) < 0)
+        goto drop_failed;
 
-      if (pa_stream_drop (pulsesrc->stream) < 0) {
-        GST_ELEMENT_ERROR (pulsesrc, RESOURCE, FAILED,
-            ("pa_stream_drop() failed: %s",
-                pa_strerror (pa_context_errno (pulsesrc->context))), (NULL));
-        goto unlock_and_fail;
-      }
-
+      /* reset pointer to data */
       pulsesrc->read_buffer = NULL;
       pulsesrc->read_buffer_length = 0;
     }
   }
 
+  pulsesrc->in_read = FALSE;
   pa_threaded_mainloop_unlock (pulsesrc->mainloop);
 
   return sum;
 
   /* ERRORS */
+was_paused:
+  {
+    GST_LOG_OBJECT (pulsesrc, "we are paused");
+    goto unlock_and_fail;
+  }
+peek_failed:
+  {
+    GST_ELEMENT_ERROR (pulsesrc, RESOURCE, FAILED,
+        ("pa_stream_peek() failed: %s",
+            pa_strerror (pa_context_errno (pulsesrc->context))), (NULL));
+    goto unlock_and_fail;
+  }
+drop_failed:
+  {
+    GST_ELEMENT_ERROR (pulsesrc, RESOURCE, FAILED,
+        ("pa_stream_drop() failed: %s",
+            pa_strerror (pa_context_errno (pulsesrc->context))), (NULL));
+    goto unlock_and_fail;
+  }
 unlock_and_fail:
   {
+    pulsesrc->in_read = FALSE;
     pa_threaded_mainloop_unlock (pulsesrc->mainloop);
-    return -1;
+
+    return (guint) - 1;
   }
 }
 
+/* return the delay in samples */
 static guint
 gst_pulsesrc_delay (GstAudioSrc * asrc)
 {
-  GstPulseSrc *pulsesrc = GST_PULSESRC (asrc);
-
+  GstPulseSrc *pulsesrc = GST_PULSESRC_CAST (asrc);
   pa_usec_t t;
-
-  int negative;
+  int negative, res;
+  guint result;
 
   pa_threaded_mainloop_lock (pulsesrc->mainloop);
+  if (gst_pulsesrc_is_dead (pulsesrc))
+    goto server_dead;
 
-  CHECK_DEAD_GOTO (pulsesrc, unlock_and_fail);
-
-  if (pa_stream_get_latency (pulsesrc->stream, &t, &negative) < 0) {
-
-    if (pa_context_errno (pulsesrc->context) != PA_ERR_NODATA) {
-      GST_ELEMENT_ERROR (pulsesrc, RESOURCE, FAILED,
-          ("pa_stream_get_latency() failed: %s",
-              pa_strerror (pa_context_errno (pulsesrc->context))), (NULL));
-      goto unlock_and_fail;
-    }
-
-    GST_WARNING_OBJECT (pulsesrc, "Not data while querying latency");
-    t = 0;
-  } else if (negative)
-    t = 0;
+  /* get the latency, this can fail when we don't have a latency update yet.
+   * We don't want to wait for latency updates here but we just return 0. */
+  res = pa_stream_get_latency (pulsesrc->stream, &t, &negative);
 
   pa_threaded_mainloop_unlock (pulsesrc->mainloop);
 
-  return (guint) ((t * pulsesrc->sample_spec.rate) / 1000000LL);
+  if (res > 0) {
+    GST_DEBUG_OBJECT (pulsesrc, "could not get latency");
+    result = 0;
+  } else {
+    if (negative)
+      result = 0;
+    else
+      result = (guint) ((t * pulsesrc->sample_spec.rate) / 1000000LL);
+  }
+  return result;
 
-unlock_and_fail:
-
-  pa_threaded_mainloop_unlock (pulsesrc->mainloop);
-  return 0;
+  /* ERRORS */
+server_dead:
+  {
+    GST_DEBUG_OBJECT (pulsesrc, "the server is dead");
+    pa_threaded_mainloop_unlock (pulsesrc->mainloop);
+    return 0;
+  }
 }
 
 static gboolean
@@ -649,12 +812,8 @@ gst_pulsesrc_create_stream (GstPulseSrc * pulsesrc, GstCaps * caps)
 
   pa_threaded_mainloop_lock (pulsesrc->mainloop);
 
-  if (!pulsesrc->context
-      || pa_context_get_state (pulsesrc->context) != PA_CONTEXT_READY) {
-    GST_ELEMENT_ERROR (pulsesrc, RESOURCE, FAILED, ("Bad context state: %s",
-            pulsesrc->
-            context ? pa_strerror (pa_context_errno (pulsesrc->context)) :
-            NULL), (NULL));
+  if (!pulsesrc->context) {
+    GST_ELEMENT_ERROR (pulsesrc, RESOURCE, FAILED, ("Bad context"), (NULL));
     goto unlock_and_fail;
   }
 
@@ -692,10 +851,20 @@ gst_pulsesrc_create_stream (GstPulseSrc * pulsesrc, GstCaps * caps)
       pulsesrc);
   pa_stream_set_read_callback (pulsesrc->stream, gst_pulsesrc_stream_request_cb,
       pulsesrc);
+  pa_stream_set_underflow_callback (pulsesrc->stream,
+      gst_pulsesrc_stream_underflow_cb, pulsesrc);
+  pa_stream_set_overflow_callback (pulsesrc->stream,
+      gst_pulsesrc_stream_overflow_cb, pulsesrc);
+  pa_stream_set_latency_update_callback (pulsesrc->stream,
+      gst_pulsesrc_stream_latency_update_cb, pulsesrc);
+
+  pa_threaded_mainloop_unlock (pulsesrc->mainloop);
 
   return TRUE;
 
 unlock_and_fail:
+  gst_pulsesrc_destroy_stream (pulsesrc);
+
   pa_threaded_mainloop_unlock (pulsesrc->mainloop);
 
 fail:
@@ -755,10 +924,9 @@ gst_pulsesrc_negotiate (GstBaseSrc * basesrc)
         result = TRUE;
       } else if (gst_caps_is_fixed (caps)) {
         /* yay, fixed caps, use those then */
-        result = gst_pulsesrc_create_stream (GST_PULSESRC (basesrc), caps);
+        result = gst_pulsesrc_create_stream (GST_PULSESRC_CAST (basesrc), caps);
         if (result)
-          gst_pad_set_caps (GST_BASE_SRC_PAD (basesrc), caps);
-        result = TRUE;
+          result = gst_pad_set_caps (GST_BASE_SRC_PAD (basesrc), caps);
       }
     }
     gst_caps_unref (caps);
@@ -778,70 +946,268 @@ static gboolean
 gst_pulsesrc_prepare (GstAudioSrc * asrc, GstRingBufferSpec * spec)
 {
   pa_buffer_attr buf_attr;
-  GstPulseSrc *pulsesrc = GST_PULSESRC (asrc);
+  const pa_buffer_attr *buf_attr_ptr;
+  GstPulseSrc *pulsesrc = GST_PULSESRC_CAST (asrc);
 
-  memset (&buf_attr, 0, sizeof (buf_attr));
-  buf_attr.maxlength = spec->segtotal * spec->segsize * 2;
+  pa_threaded_mainloop_lock (pulsesrc->mainloop);
+
+  buf_attr.maxlength = -1;
+  buf_attr.tlength = -1;
+  buf_attr.prebuf = 0;
+  buf_attr.minreq = -1;
   buf_attr.fragsize = spec->segsize;
 
+  GST_INFO_OBJECT (pulsesrc, "maxlength: %d", buf_attr.maxlength);
+  GST_INFO_OBJECT (pulsesrc, "tlength:   %d", buf_attr.tlength);
+  GST_INFO_OBJECT (pulsesrc, "prebuf:    %d", buf_attr.prebuf);
+  GST_INFO_OBJECT (pulsesrc, "minreq:    %d", buf_attr.minreq);
+  GST_INFO_OBJECT (pulsesrc, "fragsize:  %d", buf_attr.fragsize);
+
   if (pa_stream_connect_record (pulsesrc->stream, pulsesrc->device, &buf_attr,
-          PA_STREAM_INTERPOLATE_TIMING | PA_STREAM_AUTO_TIMING_UPDATE |
-          PA_STREAM_NOT_MONOTONOUS) < 0) {
+          PA_STREAM_INTERPOLATE_TIMING |
+          PA_STREAM_AUTO_TIMING_UPDATE | PA_STREAM_NOT_MONOTONOUS |
+#if HAVE_PULSE_0_9_11
+          PA_STREAM_ADJUST_LATENCY |
+#endif
+          PA_STREAM_START_CORKED) < 0) {
     GST_ELEMENT_ERROR (pulsesrc, RESOURCE, FAILED,
         ("Failed to connect stream: %s",
             pa_strerror (pa_context_errno (pulsesrc->context))), (NULL));
     goto unlock_and_fail;
   }
 
-  /* Wait until the stream is ready */
-  pa_threaded_mainloop_wait (pulsesrc->mainloop);
+  pulsesrc->corked = TRUE;
 
-  if (pa_stream_get_state (pulsesrc->stream) != PA_STREAM_READY) {
-    GST_ELEMENT_ERROR (pulsesrc, RESOURCE, FAILED,
-        ("Failed to connect stream: %s",
-            pa_strerror (pa_context_errno (pulsesrc->context))), (NULL));
-    goto unlock_and_fail;
+  for (;;) {
+    pa_stream_state_t state;
+
+    state = pa_stream_get_state (pulsesrc->stream);
+
+    if (!PA_STREAM_IS_GOOD (state)) {
+      GST_ELEMENT_ERROR (pulsesrc, RESOURCE, FAILED,
+          ("Failed to connect stream: %s",
+              pa_strerror (pa_context_errno (pulsesrc->context))), (NULL));
+      goto unlock_and_fail;
+    }
+
+    if (state == PA_STREAM_READY)
+      break;
+
+    /* Wait until the stream is ready */
+    pa_threaded_mainloop_wait (pulsesrc->mainloop);
   }
 
+  /* get the actual buffering properties now */
+  buf_attr_ptr = pa_stream_get_buffer_attr (pulsesrc->stream);
+
+  GST_INFO_OBJECT (pulsesrc, "maxlength: %d", buf_attr_ptr->maxlength);
+  GST_INFO_OBJECT (pulsesrc, "tlength:   %d (wanted: %d)",
+      buf_attr_ptr->tlength, buf_attr.tlength);
+  GST_INFO_OBJECT (pulsesrc, "prebuf:    %d", buf_attr_ptr->prebuf);
+  GST_INFO_OBJECT (pulsesrc, "minreq:    %d (wanted %d)", buf_attr_ptr->minreq,
+      buf_attr.minreq);
+  GST_INFO_OBJECT (pulsesrc, "fragsize:  %d (wanted %d)",
+      buf_attr_ptr->fragsize, buf_attr.fragsize);
+
+  /* adjust latency again */
+  spec->segsize = buf_attr_ptr->fragsize;
+  spec->segtotal = buf_attr_ptr->maxlength / spec->segsize;
+
   pa_threaded_mainloop_unlock (pulsesrc->mainloop);
 
   return TRUE;
 
 unlock_and_fail:
+  {
+    gst_pulsesrc_destroy_stream (pulsesrc);
+
+    pa_threaded_mainloop_unlock (pulsesrc->mainloop);
+    return FALSE;
+  }
+}
+
+static void
+gst_pulsesrc_success_cb (pa_stream * s, int success, void *userdata)
+{
+  GstPulseSrc *pulsesrc = GST_PULSESRC_CAST (userdata);
+
+  pulsesrc->operation_success = !!success;
+  pa_threaded_mainloop_signal (pulsesrc->mainloop, 0);
+}
+
+static void
+gst_pulsesrc_reset (GstAudioSrc * asrc)
+{
+  GstPulseSrc *pulsesrc = GST_PULSESRC_CAST (asrc);
+  pa_operation *o = NULL;
+
+  pa_threaded_mainloop_lock (pulsesrc->mainloop);
+  GST_DEBUG_OBJECT (pulsesrc, "reset");
+
+  if (gst_pulsesrc_is_dead (pulsesrc))
+    goto unlock_and_fail;
+
+  if (!(o =
+          pa_stream_flush (pulsesrc->stream, gst_pulsesrc_success_cb,
+              pulsesrc))) {
+    GST_ELEMENT_ERROR (pulsesrc, RESOURCE, FAILED,
+        ("pa_stream_flush() failed: %s",
+            pa_strerror (pa_context_errno (pulsesrc->context))), (NULL));
+    goto unlock_and_fail;
+  }
+
+  pulsesrc->paused = TRUE;
+  /* Inform anyone waiting in _write() call that it shall wakeup */
+  if (pulsesrc->in_read) {
+    pa_threaded_mainloop_signal (pulsesrc->mainloop, 0);
+  }
+
+  pulsesrc->operation_success = FALSE;
+  while (pa_operation_get_state (o) == PA_OPERATION_RUNNING) {
+
+    if (gst_pulsesrc_is_dead (pulsesrc))
+      goto unlock_and_fail;
+
+    pa_threaded_mainloop_wait (pulsesrc->mainloop);
+  }
+
+  if (!pulsesrc->operation_success) {
+    GST_ELEMENT_ERROR (pulsesrc, RESOURCE, FAILED, ("Flush failed: %s",
+            pa_strerror (pa_context_errno (pulsesrc->context))), (NULL));
+    goto unlock_and_fail;
+  }
+
+unlock_and_fail:
+
+  if (o) {
+    pa_operation_cancel (o);
+    pa_operation_unref (o);
+  }
+
   pa_threaded_mainloop_unlock (pulsesrc->mainloop);
-  return FALSE;
+}
+
+/* update the corked state of a stream, must be called with the mainloop
+ * lock */
+static gboolean
+gst_pulsesrc_set_corked (GstPulseSrc * psrc, gboolean corked, gboolean wait)
+{
+  pa_operation *o = NULL;
+  gboolean res = FALSE;
+
+  GST_DEBUG_OBJECT (psrc, "setting corked state to %d", corked);
+  if (psrc->corked != corked) {
+    if (!(o = pa_stream_cork (psrc->stream, corked,
+                gst_pulsesrc_success_cb, psrc)))
+      goto cork_failed;
+
+    while (wait && pa_operation_get_state (o) == PA_OPERATION_RUNNING) {
+      pa_threaded_mainloop_wait (psrc->mainloop);
+      if (gst_pulsesrc_is_dead (psrc))
+        goto server_dead;
+    }
+    psrc->corked = corked;
+  } else {
+    GST_DEBUG_OBJECT (psrc, "skipping, already in requested state");
+  }
+  res = TRUE;
+
+cleanup:
+  if (o)
+    pa_operation_unref (o);
+
+  return res;
+
+  /* ERRORS */
+server_dead:
+  {
+    GST_DEBUG_OBJECT (psrc, "the server is dead");
+    goto cleanup;
+  }
+cork_failed:
+  {
+    GST_ELEMENT_ERROR (psrc, RESOURCE, FAILED,
+        ("pa_stream_cork() failed: %s",
+            pa_strerror (pa_context_errno (psrc->context))), (NULL));
+    goto cleanup;
+  }
+}
+
+/* start/resume playback ASAP */
+static gboolean
+gst_pulsesrc_play (GstPulseSrc * psrc)
+{
+  pa_threaded_mainloop_lock (psrc->mainloop);
+  GST_DEBUG_OBJECT (psrc, "playing");
+  psrc->paused = FALSE;
+  gst_pulsesrc_set_corked (psrc, FALSE, FALSE);
+  pa_threaded_mainloop_unlock (psrc->mainloop);
+
+  return TRUE;
+}
+
+/* pause/stop playback ASAP */
+static gboolean
+gst_pulsesrc_pause (GstPulseSrc * psrc)
+{
+  pa_threaded_mainloop_lock (psrc->mainloop);
+  GST_DEBUG_OBJECT (psrc, "pausing");
+  /* make sure the commit method stops writing */
+  psrc->paused = TRUE;
+  if (psrc->in_read) {
+    /* we are waiting in a read, signal */
+    GST_DEBUG_OBJECT (psrc, "signal read");
+    pa_threaded_mainloop_signal (psrc->mainloop, 0);
+  }
+  pa_threaded_mainloop_unlock (psrc->mainloop);
+
+  return TRUE;
 }
 
 static GstStateChangeReturn
 gst_pulsesrc_change_state (GstElement * element, GstStateChange transition)
 {
-  GstPulseSrc *this = GST_PULSESRC (element);
+  GstStateChangeReturn ret;
+  GstPulseSrc *this = GST_PULSESRC_CAST (element);
 
   switch (transition) {
     case GST_STATE_CHANGE_NULL_TO_READY:
-
       if (!this->mixer)
         this->mixer =
             gst_pulsemixer_ctrl_new (G_OBJECT (this), this->server,
             this->device, GST_PULSEMIXER_SOURCE);
-
       break;
+    case GST_STATE_CHANGE_PAUSED_TO_PLAYING:
+      /* uncork and start recording */
+      gst_pulsesrc_play (this);
+      break;
+    case GST_STATE_CHANGE_PLAYING_TO_PAUSED:
+      /* stop recording ASAP by corking */
+      pa_threaded_mainloop_lock (this->mainloop);
+      GST_DEBUG_OBJECT (this, "corking");
+      gst_pulsesrc_set_corked (this, TRUE, FALSE);
+      pa_threaded_mainloop_unlock (this->mainloop);
+      break;
+    default:
+      break;
+  }
 
+  ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);
+
+  switch (transition) {
+    case GST_STATE_CHANGE_PLAYING_TO_PAUSED:
+      /* now make sure we get out of the _read method */
+      gst_pulsesrc_pause (this);
+      break;
     case GST_STATE_CHANGE_READY_TO_NULL:
-
       if (this->mixer) {
         gst_pulsemixer_ctrl_free (this->mixer);
         this->mixer = NULL;
       }
-
       break;
-
     default:
-      ;
+      break;
   }
 
-  if (GST_ELEMENT_CLASS (parent_class)->change_state)
-    return GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);
-
-  return GST_STATE_CHANGE_SUCCESS;
+  return ret;
 }
